/* ***** BEGIN LICENSE BLOCK *****
 *   Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is bulkfilters.
 *
 * The Initial Developer of the Original Code is
 * Mozilla Messaging, Inc.
 * Portions created by the Initial Developer are Copyright (C) 2008
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Shane Caraveo <shanec@mozillamessaging.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

Components.utils.import("resource://gre/modules/iteratorUtils.jsm");
Components.utils.import("resource://app/modules/virtualFolderWrapper.js");

var bulkFilter= {};
(function()
{
const EXT_ID = "bulkfilter@mozillamesssaging.com";

var logger = Components.classes["@mozilla.org/consoleservice;1"]
    .getService(Components.interfaces.nsIConsoleService);
var prefs = Components.classes["@mozilla.org/preferences-service;1"]
                     .getService(Components.interfaces.nsIPrefBranch);

function createSearchTermItem(type, searchValue) {
    let attrib = type == "address" ? Components.interfaces.nsMsgSearchAttrib.AllAddresses :
                                     Components.interfaces.nsMsgSearchAttrib.Subject
    return {
            value: searchValue,
            op: Components.interfaces.nsMsgSearchOp.Contains,
            attrib: attrib,
            matchAll: false,
            booleanAnd: false
            }
}

let defaultFilters = {
    "eBay":  {
        "type": "Notifications",
        "terms": [["address","@ebay"]]
    },
    "PayPal":{
        "type": "Notifications",
        "terms": [["address","paypal.com"]]
    },
    "EFF":{
        "type": "Notifications",
        "terms": [["address","editor@eff.org"]]
    },
    "Facebook":{
        "type": "Notifications",
        "terms": [["address","facebookmail.com"],
                ["address","facebookappmail.com"]]
    },
    "CouchSurfing":{
        "type": "Notifications",
        "terms": [["address","couchsurfing.com"],
                  ["address","couchsurfing.org"]]
    },
    "Wordpress":{
        "type": "Notifications",
        "terms": [["address","@wordpress.com"]]
    },
    "Twitter":{
        "type": "Notifications",
        "terms": [["address","@postmaster.twitter.com"]]
    },
    "LinkedIn":{
        "type": "Notifications",
        "terms": [["address","linkedin.com"],
                ["subject","linkedin"]
                ]
    },
    "Amazon":{
        "type": "Notifications",
        "terms": [["address","@amazon.com"]]
    },
    "Google Alerts":{
        "type": "Notifications",
        "terms": [["address","googlealerts-noreply@google.com"],
                ["address","calendar-notification@google.com"]
                ]
    },
    "Travel":{
        "type": "Notifications",
        "terms": [
                ["address","airline"],
                ["subject","airline"],
                ["address","hotel"],
                ["subject","hotel"],
                ["address","hostel"],
                ["subject","hostel"],
                ["address","hostal"],
                ["subject","hostal"],
                ["address","delta.com"],
                ["address","aa.com"],
                ["address","continental.com"],
                ["address","united.com"],
                ["address","klm.com"],
                ["address","airfrance.com"],
                ["address","britishairways.com"],
                ["address","usairways.com"],
                ["address","@allegiantair.us"],
                ["address","@mymileageplan.com"],
                ["address","clickair.com"],
                ["address","vueling.com"],
                ["address","alaskaair.com"],
                ["address","aircanada.com"],
                ["address","westjet.com"],
                ["address","ryanair"],
                ["address","travelocity.com"],
                ["address","exitnowmarketing.com"],
                ["address","transatmemberservice.com"]
            ]
    },
    "Plaxo":{
        "type": "Notifications",
        "terms": [["address","plaxo.com"]]
    }
};

/**
 * a list of search terms we want to use for our inbox search filter
 * discover data, readable name: type and list of search terms
 */
var bulklists = {};

try {
    /* this depends on a patch landing for folderPane.js,
       thunderbird bug 553747 */
    let folderTypes = {};
    for each(let [name, data] in Iterator(bulklists)) {
        folderTypes[data.type]=1;
    }
    for each(let [folderType,] in Iterator(folderTypes)) {
        gFolderTreeView.getFolderTreeMode("smart")
            .addSmartFolderType(folderType, false, false);
    }
} catch(e) {
    /* smartfolder addon patch not applied */
}

/**
 * a custom search action that allows us to run a bit of javascript to
 * handle the action.
 * implements nsIMsgFilterCustomAction
 */
var bulkSearchFolderAction = {
    id: "maillistfilter@mozillamessaging.com#bulkFolderAction",
    name: "bulkFolderAction",
    isValidForType: function bulkSearchFolderAction_isValidForType(type, scope) {
        //logger.logStringMessage("bulkFilter.valid for type?");
        return type == Components.interfaces.nsMsgFilterType.Inbox |
                    Components.interfaces.nsMsgFilterType.PostPlugin;
    },
    validateActionValue: function bulkSearchFolderAction_validateActionValue(actionValue, actionFolder, filterType) {
        //logger.logStringMessage("bulkFilter.validateActionValue "+actionValue+" on "+actionFolder.name);
        if (!bulklists[actionValue])
            return "Bulk list is not defined";
        return null;
    },
    allowDuplicates: false,
    isAsync: false,
    apply: function bulkSearchFolderAction_apply(msgHdrs, actionValue, copyListener, filterType, msgWindow) {
        //logger.logStringMessage("bulkFilter.action "+actionValue);
        try {
            // assuming all msgHdrs are in the same account and folder for now
            var msgHdr = msgHdrs.queryElementAt(0, Components.interfaces.nsIMsgDBHdr);
            createBulkSearchFolder(msgHdr, actionValue);
        } catch(e) {
        logger.logStringMessage("bulkFilter.action "+e);
        }
    }
};
var filterService = Components.classes["@mozilla.org/messenger/services/filters;1"]
                    .getService(Components.interfaces.nsIMsgFilterService);
filterService.addCustomAction(bulkSearchFolderAction);

/**
 * utility function to recursivly gather subfolders into an array
 * @param folder     {nsIMsgFolder}
 * @param outfolders {hash} to store found folders in
 */
function addSubFolders(folder, outFolders) {
  var subFolders = folder.subFolders;
  while (subFolders.hasMoreElements()) {
    var nextFolder =
      subFolders.getNext().QueryInterface(Components.interfaces.nsIMsgFolder);

    if (!(nextFolder.flags & Components.interfaces.nsMsgFolderFlags.Virtual)) {
      if (!nextFolder.noSelect)
        outFolders.push(nextFolder);

      addSubFolders(nextFolder, outFolders);
    }
  }
}

/**
 * utility function to find folders/subfolders by flag, if subfolders is true,
 * add all non-virtual subfolders.
 * 
 * @param rootFolder  {nsiMsgFolder}     where to start looking
 * @param flags       {nsMsgFolderFlags} what flags to look for
 * @param outfolders  {hash}             hash to store found folders in
 * @param subfolders  {boolean}          flag to search subfolders
 */
function addFoldersWithFlags(rootFolder, flags, outfolders, subfolders) {
  let folders = rootFolder.getFoldersWithFlags(flags);
  if (folders) {
    for each(let folder in fixIterator(folders.enumerate(), Components.interfaces.nsIMsgFolder)) {
      outfolders.push(folder);
      if (subfolders)
        addSubFolders(folder, outfolders);
    }
  }
}

/**
 * Used by our custom search action to create a virtual search folder.
 * @param msgHdr      {nsIMsgDBHeader}
 * @param actionValue {string}
 */
function createBulkSearchFolder(msgHdr, actionValue) {
    let folders = [msgHdr.folder];
    let rootFolder = msgHdr.folder.server.rootFolder;

    if (!(msgHdr.folder.flags & Components.interfaces.nsMsgFolderFlags.SentMail))
        addFoldersWithFlags(rootFolder, Components.interfaces.nsMsgFolderFlags.SentMail, folders, false);
    if (!(msgHdr.folder.flags & Components.interfaces.nsMsgFolderFlags.Inbox))
        addFoldersWithFlags(rootFolder, Components.interfaces.nsMsgFolderFlags.Inbox, folders, false);
    if (!(msgHdr.folder.flags & Components.interfaces.nsMsgFolderFlags.Archive))
        addFoldersWithFlags(rootFolder, Components.interfaces.nsMsgFolderFlags.Archive, folders, true);

    // create the search folder, actionValue is the key into bulklists
    let terms = [];
    let termCreator = Components.classes["@mozilla.org/messenger/searchSession;1"]
                             .createInstance(Components.interfaces.nsIMsgSearchSession);
    let data = bulklists[actionValue];
    for each(let termData in data.terms) {
        terms.push(getTermForSearch(termData, termCreator));
    }
    
    let baseFolder;
    try {
        baseFolder = rootFolder.getChildNamed(data.type);
    } catch(e) {
        baseFolder = rootFolder.addSubfolder(data.type);
        baseFolder.setFlag(nsMsgFolderFlags.Virtual);
        baseFolder.setStringProperty("customViewURL", "extensions."+EXT_ID+".folderview.url")
        baseFolder.setStringProperty("addon", EXT_ID)
    }
    let searchFolder;
    try {
        searchFolder = baseFolder.getChildNamed(actionValue);
    } catch(e) {}
    if (!searchFolder) {
        // create the virtual search folder for this
        searchFolder = VirtualFolderHelper.createNewVirtualFolder(
                                     actionValue,
                                     baseFolder,
                                     folders,
                                     terms,
                                     false /*onlineSearch*/);
        searchFolder.virtualFolder.setStringProperty("smartFolderName", data.type);
    } else {
        // fixup the search terms and folders in case we've updated.
        let vfolder = VirtualFolderHelper.wrapVirtualFolder(searchFolder);
        if (vfolder) {
            // just replace the search terms
            vfolder.searchTerms = terms;
        }
    }
};

/**
 * creates a search term from our configuration data
 * @param searchData {object}
 * @param termCreator {nsIMsgFilter|nsIMsgSearchSession}
 * @returns nsIMsgSearchTerm
 */
function getTermForSearch(searchData, termCreator) {
    let searchTerm = termCreator.createTerm();
    let value = searchTerm.value;
    value.str = searchData.value;
    searchTerm.value = value;
    searchTerm.attrib = searchData.attrib;
    searchTerm.op = searchData.op;
    searchTerm.matchAll = searchData.matchAll;
    searchTerm.booleanAnd = searchData.booleanAnd;
    return searchTerm;
};

var _filters = {};

/**
 * create and apply a temporary filterlist, with our configured search terms,
 * and set it on the provided folder.  Used to set our search terms onto
 * folders we care about so we can discover emails we want to create special
 * folders for.
 * @param rootFolder {nsIMsgFolder}
 */
function setFilters(rootFolder) {
    let folder = rootFolder.getFolderWithFlags(Components.interfaces.nsMsgFolderFlags.Inbox);
    let filterService =
      Components.classes["@mozilla.org/messenger/services/filters;1"]
        .getService(Components.interfaces.nsIMsgFilterService);
    let filterList = filterService.getTempFilterList(folder);

    // save for later removal
    _filters[folder.URI] = filterList;

    for each(let [name, data] in Iterator(bulklists)) {
        //logger.logStringMessage("bulkFilter create filter "+name+" for "+folder.URI);
        let filter = filterList.createFilter(name);
        for each(let searchData in data.terms) {
            filter.appendTerm(getTermForSearch(searchData, filter));
        }
        
        // our action needs to be javascript
        var action = filter.createAction();
        action.type = Components.interfaces.nsMsgFilterAction.Custom;
        action.customId = bulkSearchFolderAction.id;
        action.strValue = name;
        filter.appendAction(action);
        filter.temporary = true;
        filter.enabled = true;
        filter.filterType = Components.interfaces.nsMsgFilterType.Manual | Components.interfaces.nsMsgFilterType.PostPlugin;

        filterList.insertFilterAt(0, filter);
    }
    applyFiltersToFolder(folder, filterList)
};

/**
 * apply a filterlist to a folder
 * @param folder {nsIMsgFolder}
 * @param filterList {nsIMsgFilterList}
 */
function applyFiltersToFolder(folder, filterList) {
    //logger.logStringMessage("bulkFilter apply filters to "+folder.name);
    var folders = Components.classes["@mozilla.org/supports-array;1"].createInstance(Components.interfaces.nsISupportsArray);
    folders.AppendElement(folder);
    let filterService =
      Components.classes["@mozilla.org/messenger/services/filters;1"]
        .getService(Components.interfaces.nsIMsgFilterService);
    filterService.applyFiltersToFolders(filterList, folders, null);
};

/**
 * while folder.getStringProperty should be enough, it is not working on the
 * nightly 3.1 for me.
 */
function dbFolderProperty(aFolder, aProp) {
    let msgDatabase = aFolder.msgDatabase;
    if (msgDatabase && msgDatabase.dBFolderInfo) {
      return msgDatabase.dBFolderInfo.getCharProperty(aProp);
    }
    return null;
}

/**
 * a select handler for the folder tree so we can display our folder view
 * when our folders are selected.
 */
function folderTreeSelectHandler(event) {
  let folder = gFolderDisplay.displayedFolder;
  let addon = dbFolderProperty(folder, "addon");
  if (addon == EXT_ID) {
    var prefName = dbFolderProperty(folder,"customViewURL");
    var customViewURL = prefs.getCharPref(prefName);
    window.frames["accountCentralPane"].location.href = customViewURL;

    var accountBox = document.getElementById("accountCentralBox");
    document.getElementById("displayDeck").selectedPanel = accountBox;

    if (gFolderDisplay._tabInfo)
      mailTabType._setPaneStates(gFolderDisplay._tabInfo.mode.accountCentralLegalPanes,
        {folder: gFolderDisplay.folderPaneVisible});
  }
}

function configureBulkFilters(jsonData) {
    // read our json file, create the filters we'll use
    for each(let [name, data] in Iterator(jsonData)) {
        let list = bulklists[name] = {};
        list.type = data.type;
        list.terms = [];
        for (var i = 0; i < data.terms.length; i++) {
            let term = data.terms[i];
            list.terms.push(createSearchTermItem(term[0], term[1]));
        }
    }
}

function setupBulkFilters() {
    // get all the pop/imap servers and set our filters on them
    let acctMgr = Components.classes["@mozilla.org/messenger/account-manager;1"]
                     .getService(Components.interfaces.nsIMsgAccountManager);
    let accounts = [a for each
                    (a in fixIterator(acctMgr.accounts, Components.interfaces.nsIMsgAccount))];
    accounts = accounts.filter(function fix(a) { return a.incomingServer &&
                                                       (a.incomingServer.type == 'imap' ||
                                                        a.incomingServer.type == 'pop3'); });
    for each(let [,acct] in Iterator(accounts)) {
        setFilters(acct.incomingServer.rootFolder);
    }
    
    // override the folder view for our custom folders
    let viewurl;
    try {
        viewurl = Application.extensions.get(EXT_ID).prefs
                             .getValue("folderview.url", null);
    } catch(e) {
            logger.logStringMessage("bulkFilter.onLoad "+e);

    }
    if (viewurl) {
      window.document.getElementById('folderTree').
        addEventListener("select", folderTreeSelectHandler, false);
    }     
}

/**
 * when the window is loaded, create our search filters and apply them.
 * discovery will cause our special folders to be created if necessary.
 */
this.onLoad = function bulkFilter_onLoad() {
    //logger.logStringMessage("bulkFilter.onLoad");
    try {
        let targetURL = "http://people.mozillamessaging.com/shanec/bulkfilters.json";
        let req = Components.classes["@mozilla.org/xmlextras/xmlhttprequest;1"].createInstance(Components.interfaces.nsIXMLHttpRequest);  
        //logger.logStringMessage("REQUEST: "+targetURL);

        let self = this;
        req.open('GET', targetURL, true);
        req.onreadystatechange = function (aEvt) {
            if (req.readyState == 4) {
                if (req.status == 200) {
                    configureBulkFilters(JSON.parse(req.responseText));
                    setupBulkFilters();
                } else {
                    // set SOMETHING up
                    //logger.logStringMessage("bulkFilter.ERROR loading json config, using defaults");
                    configureBulkFilters(defaultFilters);
                    setupBulkFilters();
                }
            }
        }
        req.send(null);
    } catch(e) {
        logger.logStringMessage("bulkFilter.onLoad "+e);
    }
};

}).apply(bulkFilter);


window.addEventListener("load", function(e) { bulkFilter.onLoad(e); }, false);
